/****************** AHB Interface PSL Sugar Assertions ****************/
`timescale 1 ns / 100 ps
module ahbCompliance 
  (
   //	  hAddr_i, hTrans_i, hWrite_i, hSize_i,
   //	  hBurst_i, hProt_i, hWData_i, hRData_i, hReady_i, hResp_i,
   //	  hReq_i, hLock_i, hGnt_i, hResetn_i, hSel_i, hMaster_i,
   //	  hMastLock_i, hSplit_i, hClk_i;
   input hclk,         // Bus clock,  rising edge of HCLK.
   input hresetn,      // Reset  active LOW 
   input [31:0] haddr,  // Address bus, Master 
   input [1:0] htrans,  // Master Transfer type,  NONSEQUENTIAL, SEQUENTIAL, IDLE or BUSY.
   input hwrite,       // Transfer direction, HIGH =>  write; LOW => read 
   input [2:0] hsize,   // Transfer size, byte (8-bit), halfword (16-bit) or word (32-bit).
   input [2:0] hburst,  // Burst type, Four, eight and sixteen beat, incrementing or wrapping.
   input [3:0] hprot,   // Protection control
   
   input [31:0] hwdata, // master Write data bus
   input hsel,        // Slave select
   input [31:0] hrdata, // Slave Read data bus
   input hready,       // Slave Transfer done, HIGH => finished; LOW => extend transfer.
   input [1:0] hresp);  // Transfer response: OKAY, ERROR, RETRY and SPLIT.
  
  localparam dataBusWidth = 32;
  localparam numSlaves = 32;
  localparam busMaster = 0;
  /* these localparams define the Transfer Type as specified by HTrans */
  localparam IDLE = 2'b00;
  localparam BUSY = 2'b01;
  localparam SEQ = 2'b11;
  localparam NONSEQ = 2'b10;
  /* these localparams define the Burst Mode as specified by HBurst */
  localparam SINGLE = 3'b000;
  localparam INCR = 3'b001;
  localparam WRAP4 = 3'b010;
  localparam INCR4 = 3'b011;
  localparam WRAP8 = 3'b100;
  localparam INCR8 = 3'b101;
  localparam WRAP16 = 3'b110;
  localparam INCR16 = 3'b111;
  /* define the amount of data in each beat as specified by HSize*/
  localparam bits8 = 3'b000;
  localparam bits16 = 3'b001;
  localparam bits32 = 3'b010;
  localparam bits64 = 3'b011;
  localparam bits128 = 3'b100;
  localparam bits256 = 3'b101;
  localparam bits512 = 3'b110;
  localparam bits1024 = 3'b111;
  /* define the transfer response signals as specified by HResp*/
  localparam OKAY = 2'b00;
  localparam ERROR = 2'b01;
  localparam RETRY = 2'b10;
  localparam SPLIT = 2'b11;
  
  //input [31:0] haddr, hwdata, hrdata;
  //input [1:0]  htrans, hresp;
  //input 	   hwrite, hready, hReq;
  //input [2:0]  hsize, hburst;
  //input [3:0]  hprot, hmaster;
  //input 	   hlock, hgnt, hmastlock;
  //input [numSlaves-1:0] hsel;
  //input [15:0] 			hsplit;
  //input 				hresetn, hclk;
  
  reg [31:0] 			prev_haddr, prev_haddrPlusSize;
  reg 					prev_hwrite, prev_hready;
  reg [numSlaves-1:0] 	prev_hsel;
  reg [3:0] 			prev_hprot;
  reg [2:0] 			prev_hsize, prev_hburst;
  reg [dataBusWidth-1:0] prev_hrdata, prev_hwdata;
  reg [1:0] 			 prev_htrans, prev_hresp;
  integer 				 AddrIncrement;
  integer 				 NumberBeats=0;
  /* capture data that is needed in the following assertions */
  always @(posedge hclk or negedge hresetn)
  begin
	prev_haddr <= haddr;
	prev_hready <= hready;
	prev_hwrite <= hwrite;
	prev_hrdata <= hrdata;
	prev_hwdata <= hwdata;
	prev_htrans <= htrans;
	prev_hburst <= hburst;
	prev_hresp <= hresp;
	prev_hsel <= hsel;
	prev_hsize <= hsize;
	prev_hprot <= hprot;
	case (hsize) /* AddrIncrement is used to check the address */
	  bits8: AddrIncrement=1;
	  bits16: AddrIncrement=2;
	  bits32: AddrIncrement=4;
	  bits64: AddrIncrement=8;
	  bits128: AddrIncrement=16;
	  bits256: AddrIncrement=32;
	  bits512: AddrIncrement=64;
	  bits1024: AddrIncrement=128;
	  default: ;
	endcase
	/* NumberBeats is used to check packet lengths */
	if (htrans == IDLE)
	  NumberBeats <= 0;
	else if ( htrans == NONSEQ)
	  NumberBeats <= 1;
	else if ( (htrans == SEQ) && (hready == 1) )
	  NumberBeats <= NumberBeats + 1;
	else
	  NumberBeats <= NumberBeats;
	/* determine the predicted address */
	if ((htrans != BUSY) && (hready != 0))
	  prev_haddrPlusSize <= haddr + AddrIncrement;
  end
  /**************************** PSL ASSERTIONS **************************/
  /* define default clock that will be used to sample all subsequent asserts */
  /* NOTE: it is important that data does not change on the same edge as clock
   because, like HDL, assertions can be sensitive to races */
  // sugar default clock = (posedge hclk);
  
  /* Behavior: If reset is active then Transfer Type is IDLE and Resp is OKAY
   (default clock) */
  // sugar property IdleAndOkayDuringReset = always(
  // {(hresetn == 0);(hresetn == 1)} |->
  // {(htrans == IDLE) && (hresp == OKAY)});
  
  /* Behavior: If the transfer type is NONSEQ and Burst Mode is SINGLE,
   then on the next clock cycle, the transfer type is not SEQ and not BUSY
   (default clock) */
  // sugar property BusyAndSeqNeverFollowNonseqOnSingleTransfer = always (
  // (htrans == NONSEQ) && (hburst == SINGLE ) ->
  // next ((htrans != SEQ) && (htrans != BUSY)) );
  
  /* Behavior: If the transfer type is BUSY and slave is ready,
   then on the next clock cycle, the transfer type must not be IDLE and
   must not be NONSEQ unless grant is 0 and ready is 1 (default clock) */
  // sugar property NeverGoFromBusyToIdleOrNonseqUnlessNoGrant = always (
  // ((htrans == BUSY) && (hready == 1)) ->
  // next ((htrans != IDLE) && (htrans != NONSEQ))
  // abort ((hready == 1)) || (hresetn == 0));
  //// abort ((hgnt == 0) && (hready == 1)) || (hresetn == 0));
  
  /* Behavior: if the transfer type is IDLE, then on the next clock
   cycle, the transfer type must be either IDLE or NONSEQ (default clock)*/
  // sugar property AlwaysGoToNonseqFromIdle = always (
  // (htrans == IDLE) ->
  // next ((htrans == IDLE) || (htrans == NONSEQ)) );
  
  /* Behavior: if the transfer type is BUSY then on the corresponding
   data transfer , the slave must provide a zero wait state OKAY response
   (default clock) */
  // sugar property ResponseToBusyMustBeZeroWaitOKAY = always (
  // ((htrans == BUSY) && (hready == 1)) ->
  // next ((hresp == OKAY) && (hready == 1)) );
  
  /* Behavior: if the transfer type is IDLE then on the corresponding data
   transfer, the slave must provide a zero wait state OKAY response
   (default clock) */
  // sugar property ResponseToIdleMustBeZeroWaitOKAY = always (
  // ((htrans == IDLE) && (hready == 1)) ->
  // next ((hresp == OKAY) && (hready == 1)) );
  
  /* Behavior: if the previous response was OKAY, and the current response
   is not OKAY, then Ready must be 0 during the second not OKAY response
   (default clock) */
  // sugar property ReadyMustBeZeroDuringFirstNotOkayResponse = always (
  // {(hresp == OKAY);(hresp != OKAY)} |->
  // {hready == 0}) ;
  
  /* Behavior: if the response is SPLIT or RETRY, the master must drive IDLE
   (default clock) */
  // sugar property FirstNotOkayResponseCausesNextIdle = always (
  // ((hresp == SPLIT) || (hresp == RETRY)) && (hready == 0) ->
  // next (htrans == IDLE));
  
  /* Behavior: if the transfer type is SEQ or BUSY, then the controls of the
   corresponding data transfer must be the same as their previous control
   (default clock) NOTE: the equivalence operator is used with hprot
   because this is an optional signal and may be an "X" value if not
   connected */
  // sugar property ControlMustBeConstantDuringABurst = always (
  // (htrans == SEQ) || (htrans == BUSY) ->
  // ((hsize == prev_hsize) && (hprot === prev_hprot) &&
  // (hwrite == prev_hwrite)) );
  
  /* Behavior: if Ready is 0 and the response is OK or ERROR, the address
   and control, and data from the master are held constant (default clock)
   NOTE: the equivalence operator is used with hprot because this is an
   optional signal and may be an "X" value if not connected */
  // sugar property ControlMustBeConstantWhenSlaveNotReady = always (
  // (hready == 0) && ((hresp == OKAY) || (hresp == ERROR)) -> next
  // (htrans == prev_htrans) && (hsize == prev_hsize) &&
  // (hprot === prev_hprot) && (hwrite == prev_hwrite) &&
  // (haddr == prev_haddr) && (hburst == prev_hburst) 
  // abort (hresetn == 0));
  
  /* Behavior: if ready is 0 and the previous transfer type was SEQ or NONSEQ,
   then on the next cycle the write data is the same as the previous write data
   (default clock) (This assertion is designed to fail!!!) */
  // sugar property WriteDataMustBeConstantWhenSlaveNotReadyAndDataBeingTransferred = always (
  // (hready == 0) && (hwrite == 1) &&
  // ((prev_htrans == SEQ) || (prev_htrans == NONSEQ)) -> next
  // (hwdata == prev_hwdata) 
  // abort (hresetn == 0));
  
  /* Behavior: if transfer type is BUSY then address does not change
   (default clock) */
  // sugar property AddrHeldWhenMasterBusy = always (
  // (htrans == BUSY) -> next (haddr == prev_haddr) );
  
  /* Behavior: if transfer is in progress, then 1K boundary is not exceeded
   (default clock) */
  // sugar property PageAddressNeverExceeds1kBoundary = always (
  // ((htrans == SEQ) || (htrans == BUSY)) ->
  // (haddr[31:10] == prev_haddr[31:10]) );
  
  /* Behavior: If the burst type is not INCR and the transfer type is not
   IDLE , then the NumberBeats for the Burst Mode is not exceeded
   (default clock) */
  // sugar property BurstIsNotTooLong = always (
  // (hburst != INCR) && (htrans != IDLE) ->
  // ((hburst == SINGLE) && (NumberBeats <= 1)) ||
  // ((hburst == WRAP4) && (NumberBeats <= 4)) ||
  // ((hburst == WRAP8) && (NumberBeats <= 8)) ||
  // ((hburst == WRAP16) && (NumberBeats <= 16)) ||
  // ((hburst == INCR4) && (NumberBeats <= 4)) ||
  // ((hburst == INCR8) && (NumberBeats <= 8)) ||
  // ((hburst == INCR16) && (NumberBeats <= 16)) );
  
  /* Behavior: For all but INCR Burst mode, if the end of the packet is being
   transferred as indicated by a transition from SEQ to IDLE when Resp is ok
   then the NumberBeats for the Burst Mode is the max number unless grant is 0
   (default clock) */
  // 7/16/03 Modified to correct error in number because of last increment
  // sugar property BurstIsNotTooShort = always (
  // {((hburst != INCR) && (hresp == OKAY) && (htrans == SEQ));
  // (htrans == IDLE) } |->
  // {((prev_hburst == SINGLE) && (NumberBeats == 1)) ||
  // ((prev_hburst == WRAP4) && (NumberBeats == (4+1))) ||
  // ((prev_hburst == WRAP8) && (NumberBeats == 8)) ||
  // ((prev_hburst == WRAP16) && (NumberBeats == 16)) ||
  // ((prev_hburst == INCR4) && (NumberBeats == 4)) ||
  // ((prev_hburst == INCR8) && (NumberBeats == 8)) ||
  // ((prev_hburst == INCR16) && (NumberBeats == 16))});
  //// abort (hgnt == 0)) ;
  
  /* Behavior: if the transfer type is IDLE , then on the next clock cycle,
   the transfer type is not SEQ and the transfer type is not BUSY
   (default clock) */
  // sugar property NeverGoFromIdleToSeqOrBusy = always (
  // (htrans == IDLE) ->
  // next ((htrans != SEQ) && (htrans != BUSY)) );
  
  /* Behavior: RETRY is always asserted for two cycles unless reset is asserted
   (default clock)*/
  // sugar property RetryResponseMustPersistTwoCycles = always (
  // {(hresp != RETRY);(hresp == RETRY)} |=>
  // {(hresp == RETRY)}
  // abort (hresetn == 0));
  
  /* Behavior: SPLIT is always asserted for two cycles unless reset is asserted
   (default clock) */
  // sugar property SplitResponseMustPersistTwoCycles = always (
  // {(hresp != SPLIT);(hresp == SPLIT)} |=>
  // {(hresp == SPLIT)}
  // abort (hresetn == 0));
  
  /* Behavior: ERROR is always asserted for two cycles unless reset is asserted
   (default clock) */
  // sugar property ErrorResponseMustPersistTwoCycles = always (
  // {(hresp != ERROR);(hresp == ERROR)} |=>
  // {(hresp == ERROR)}
  // abort (hresetn == 0));
  /******* incorrectAddressCalculation for 3 example WRAP modes: This calculates
   whether the next address is correct for all wrap modes*******/
  
  /* Behavior: if Burst mode is WRAP4 and Size is 8 bits, then if
   transfer mode is SEQ or BUSY, then if previous transfer mode is not
   busy and previous Ready is not zero, then check that the address is as predicted */
  // sugar property CorrectAddressDuringPageSize4BurstWrap = always (
  // ((hburst== WRAP4) && (hsize == bits8)) ->
  // ((htrans == SEQ || htrans == BUSY)) ->
  // ((prev_htrans != BUSY) && (prev_hready != 0)) ->
  // ((haddr[31:2] == prev_haddr[31:2]) &&
  // (haddr[1:0] == prev_haddrPlusSize[1:0])) );
  //
  // sugar property CorrectAddressDuringPageSize8BurstWrap = always (
  // ((hburst== WRAP4) && (hsize == bits16)) ||
  // ((hburst== WRAP8) && (hsize == bits8)) ->
  // ((htrans == SEQ) || (htrans == BUSY)) ->
  // ((prev_htrans != BUSY) && (prev_hready != 0)) ->
  // ((haddr[31:3] == prev_haddr[31:3]) &&
  // (haddr[2:0] == prev_haddrPlusSize[2:0])) );
  //
  // sugar property CorrectAddressDuringPageSize2048BurstWrap = always (
  // ((hburst== WRAP16) && (hsize == bits1024)) ->
  // ((htrans == SEQ) || (htrans == BUSY)) ->
  // ((prev_htrans != BUSY) && (prev_hready != 0)) ->
  // ((haddr[31:11] == prev_haddr[31:11]) &&
  // (haddr[10:0] == prev_haddrPlusSize[10:0])) );
  
  /* Behavior: the address must increment by size during all INCR beats */
  // sugar property AddressIncBySizeDuringAllBurstIncrBeats = always (
  // ((hburst == INCR16) || (hburst == INCR8) ||
  // (hburst == INCR4)) ->
  // ((htrans == SEQ) || (htrans == BUSY)) ->
  // ((prev_htrans != BUSY) && (hready != 0)) ->
  // (haddr == prev_haddrPlusSize) );
  
  /* Behavior: always address must be aligned to the transfer size */
  // sugar property AddressNotAlignedToTransferSize = always (
  // ((hsize == bits8) ||
  // ((hsize == bits16) && (haddr[0] == 1'b0)) ||
  // ((hsize == bits32) && (haddr[1:0] == 2'b00)) ||
  // ((hsize == bits64) && (haddr[2:0] == 3'b000)) ||
  // ((hsize == bits128) && (haddr[3:0] == 4'b0000)) ||
  // ((hsize == bits256) && (haddr[4:0] == 5'b00000)) ||
  // ((hsize == bits512) && (haddr[5:0] == 6'b000000)) ||
  // ((hsize == bits1024) && (haddr[6:0] == 7'b0000000)) ) );
  
  /* Behavior: always the maximum number of wait states is 16 */
  // sugar property NeverMoreThan16WaitStates = always (
  // {(hready == 1);(hready == 0)} |=>
  // {{(hready == 0)[*0..15]};{hready == 1}}
  // abort (hresp != OKAY));
endmodule


